

import torch.nn as nn
from transformers import T5Tokenizer, T5EncoderModel, T5ForConditionalGeneration
from transformers import BertModel, AutoTokenizer, AutoModel, GPT2Tokenizer
#import tensorflow as tf

import pandas as pd
import numpy as np
import ast
import datasets
from datasets import load_metric
from transformers import TrainingArguments, Trainer

import pyarrow as pa
import pyarrow.dataset as ds

from torch.optim import Adam
from torch.utils.data import DataLoader
from transformers import get_scheduler, AutoModelForCausalLM, AutoConfig, AutoModelForSequenceClassification #MptForSequenceClassification

import torch
from tqdm.auto import tqdm
import statistics
import time

import subprocess as sp
import os

from sklearn.model_selection import train_test_split
import json
import random

import re
import argparse
#############################################################

def combine_query_document(query: str, document: str, answer=None):
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r"," ").replace("\t"," ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = (" ").join(cleaned_document.split(" ")[:512])

    if len(query.split(" ")) > 100:
        query = (" ").join(query.split(" ")[:30])

    if answer is None:
        return query + " | " + cleaned_document
    else:
        try:
            return query + " | " + cleaned_document + " | " + answer
        except:
            print("Error with combine_query_document")
            print("Query: " + str(query))
            print("Cleaned Document: " + str(cleaned_document))
            print("Answer: " + str(answer))
            return str(query) + " | " + str(cleaned_document) + " | " + str(answer)

def format_text_for_fine_tuning_content_relevance_sequence_classification(question: str, document: str):
    instruction = "You are an expert judge for evaluating question answering systems. "
    instruction += "Given the following question and document, you must analyze the provided document and determine whether it is sufficient for answering the question. \n\n"

    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r"," ").replace("\t"," ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = (" ").join(cleaned_document.split(" ")[:512])

    instruction += "### Instruction:\n"
    instruction += "Question: " + question + "\n"
    instruction += "Document: " + cleaned_document + "\n"
    instruction += "### Response:\n"

    return instruction

#############################################################

random_state = 42

np.random.seed(random_state)
random.seed(random_state)
torch.manual_seed(random_state)
os.environ['PYTHONHASHSEED'] = str(random_state)

############################################################

def get_gpu_memory():
    command = "nvidia-smi --query-gpu=memory.free --format=csv"
    memory_free_info = sp.check_output(command.split()).decode('ascii').split('\n')[:-1][1:]
    memory_free_values = [int(x.split()[0]) for i, x in enumerate(memory_free_info)]
    return memory_free_values

############################################################

class CustomBERTModel(nn.Module):
    def __init__(self, number_of_labels, model_choice):
          self.model_choice = model_choice
          super(CustomBERTModel, self).__init__()
          if model_choice in ["mosaicml/mpt-7b-instruct", "mosaicml/mpt-7b"]:

            config = AutoConfig.from_pretrained(model_choice, trust_remote_code=True)
            config.attn_config['attn_impl'] = 'triton'  # change this to use triton-based FlashAttention
            config.max_seq_len = max_token_length

            model_encoding = AutoModelForCausalLM.from_pretrained(
                model_choice,
                config=config,
                #max_seq_len=max_token_length,
                torch_dtype=torch.bfloat16,
                trust_remote_code=True,
                use_auth_token=True
            )
            embedding_size = 4096
            self.encoderModel = model_encoding.transformer

          elif model_choice in ['mosaicml/mpt-1b-redpajama-200b']:

            #model_encoding = AutoModelForCausalLM.from_pretrained("mosaicml/mpt-1b-redpajama-200b", trust_remote_code=True, 
            #                                                      torch_dtype=torch.bfloat16) #attn_impl='triton', 
            model_encoding = MptForSequenceClassification.from_pretrained("mosaicml/mpt-1b-redpajama-200b", trust_remote_code=True)
            embedding_size = 2048
            self.encoderModel = model_encoding
          
          elif model_choice in ["google/t5-large-lm-adapt", "google/t5-xl-lm-adapt"]:

            model_encoding = AutoModelForSequenceClassification.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding#.transformer

          elif model_choice in ["roberta-large", "microsoft/deberta-v3-large"]:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding

          elif model_choice in ["microsoft/deberta-v2-xlarge", "microsoft/deberta-v2-xxlarge"]:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1536
            self.encoderModel = model_encoding
          
          else:

            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 768
            self.encoderModel = model_encoding

          #########################################

          self.classifier = nn.Sequential(nn.Linear(embedding_size, 256), nn.Linear(256, number_of_labels))
          self.embedding_size = embedding_size

########################################################

    def forward(self, ids, mask, labels=None, decoder_input_ids=None):
        model_choice = self.model_choice # Possibly change this to self.model_choice = model_choice
        if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt", "mosaicml/mpt-1b-redpajama-200b"]:
            total_output = self.encoderModel(input_ids=ids, attention_mask=mask) #labels=labels
            return total_output['logits']
        else:
            total_output = self.encoderModel(ids, attention_mask=mask)
            sequence_output = total_output['last_hidden_state']

            last_hidden_state_formatted = sequence_output[:,0,:].view(-1, self.embedding_size)
            linear2_output = self.classifier(last_hidden_state_formatted)

            return linear2_output

############################################################

def tokenize_function(tokenizer, examples):
    return tokenizer(examples["text"], padding="max_length", truncation=True) #.input_ids

########################################################

def checkpoints(classification_dataset, model_choice):
    checkpoints_folder_path = "checkpoints/"
    if not os.path.isdir(checkpoints_folder_path):
        os.mkdir(checkpoints_folder_path)

    dataset_folder_path = "checkpoints/" + model_choice.replace("/", "-")
    if not os.path.isdir(dataset_folder_path):
        print("Creating folder: " + dataset_folder_path)
        os.mkdir(dataset_folder_path)

    for dataset in classification_datasets:
        try:
            os.mkdir(dataset_folder_path + "/" + dataset.replace("../", "").replace("/", "-"))
        except:
            print("Already exists")
            print(dataset_folder_path + "/" + dataset.replace("../", "").replace("/", "-"))

########################################################

def load_model(model_choice):
    max_token_length = 2048
    tokenizer = AutoTokenizer.from_pretrained(model_choice, model_max_length=max_token_length)
    
    return tokenizer, max_token_length

########################################################

def prepare_and_clean_data(dataset, learning_rate_choices, chosen_learning_rate, model_choice, 
                number_of_runs, validation_set_scoring, 
                label_column, validation_set, patience_value, 
                num_epochs, num_warmup_steps, gradient_accumulation_multiplier, 
                assigned_batch_size, tokenizer): 

    print("--------------------------------------------------------------------------")
    print("Starting new learning rate: " + str(chosen_learning_rate))
    print("--------------------------------------------------------------------------")

        
    from random import randrange
    random_int = randrange(1000000)

    checkpoint_path = "checkpoints/" + model_choice.replace("/", "-") + "/" + str(chosen_learning_rate) + "_"
    checkpoint_path += str(number_of_runs) + "_" + str(validation_set_scoring) + "_" + label_column + "_" + str(validation_set.split("/")[-1].replace(".tsv", "")) + "_" + str(random_int) + ".pt"

    # checkpoint_path = "checkpoints/" + model_choice.replace("/", "-") + "/" + dataset.replace("../", "").replace("/", "-") + "/" + str(chosen_learning_rate) + "_"
    # checkpoint_path += str(number_of_runs) + "_" + str(validation_set_scoring) + "_" + label_column + "_" + str(validation_set.split("/")[-1].replace(".tsv", "")) + "_" + str(random_int) + ".pt"

    # checkpoint_path = "checkpoints/" + model_choice.replace("/", "-") + "/" + dataset.replace("../", "").replace("/", "-") + "/" + str(chosen_learning_rate) + "_"
    # checkpoint_path += str(number_of_runs) + "_" + str(validation_set_scoring) + "_" + label_column + "_" + str(validation_set.split("/")[-1].replace(".tsv", "")) + "_" + str(random_int) + ".pt"
    
    execution_start = time.time()

    print("Dataset: " + dataset)
    print("Model: " + model_choice)
    print("Test Set Selection: " + validation_set)
    print("Number of Runs: " + str(number_of_runs))
    print('Learning Rate: ' + str(chosen_learning_rate))
    print("Checkpoint Path: " + checkpoint_path)
    print("Patience: " + str(patience_value))
    print("Validation Set Choice: " + str(validation_set_scoring))
    print("Number of Epochs: " + str(num_epochs))
    print("Number of warmup steps: " + str(num_warmup_steps))

    return checkpoint_path, patience_value

########################################################

def analyze_and_report_data(dataset, label_column, tokenizer, max_token_length): 

    synth_queries = pd.read_csv(dataset, sep="\t")
    if "nq_reformatted" in dataset:
        synth_queries['synthetic_query'] = synth_queries['Query']
        synth_queries['generated_answer'] = synth_queries['Answer']
        synth_queries['document'] = synth_queries['Document']

    print("Positive and Negative Label Count")
    print(synth_queries[label_column].tolist().count("Yes"))
    print(synth_queries[label_column].tolist().count("No"))
    print(set(synth_queries[label_column].tolist()))
    
    synth_queries = synth_queries[synth_queries[label_column] != "NaN"]
    synth_queries = synth_queries[synth_queries["synthetic_query"].notna()]
    synth_queries = synth_queries[synth_queries["document"].notna()]
    synth_queries = synth_queries[synth_queries['generated_answer'].notna()]
    synth_queries = synth_queries[synth_queries[label_column].notna()]
    synth_queries = synth_queries.sample(n=len(synth_queries), random_state=42)
    #synth_queries = synth_queries[:40000]

    print("Number of unique questions")
    print(len(set(synth_queries['synthetic_query'].tolist())))
    print("Positive and Negative Label Count")
    print(synth_queries[label_column].tolist().count("Yes"))
    print(synth_queries[label_column].tolist().count("No"))
    print(set(synth_queries[label_column].tolist()))
    
    if "Context" in label_column:
        synth_queries["concat_text"] = [combine_query_document(synth_queries.iloc[i]['synthetic_query'], synth_queries.iloc[i]['document']) for i in range(len(synth_queries))]
    else:
        synth_queries["concat_text"] = [combine_query_document(synth_queries.iloc[i]['synthetic_query'], synth_queries.iloc[i]['document'], synth_queries.iloc[i]['generated_answer']) for i in range(len(synth_queries))]
    synth_queries['token_length'] = [len(tokenizer.encode(synth_queries.iloc[i]['concat_text'], return_tensors='pt')[0]) for i in range(len(synth_queries))]
    synth_queries = synth_queries.drop_duplicates(["concat_text"])

    print("Max token length")
    print(synth_queries['token_length'].max())
    print("Total inputs over max token length set of " + str(max_token_length))
    print(len(synth_queries[synth_queries['token_length'] > max_token_length]))

    synth_queries = synth_queries[synth_queries['token_length'] <= 2048]

    return synth_queries

########################################################

def transform_data(synth_queries, validation_set, label_column):

    train_df = synth_queries

    test_set = pd.read_csv(validation_set, sep="\t")
    test_set['Question'] = test_set['Query']
    test_set['Document'] = test_set['Document'].str.strip()
    test_set = test_set[test_set["Document"].str.len() > 100]
    test_set = test_set[test_set[label_column].notna()]

    train_df['document'] = train_df['document'].astype(str).str.strip()
    train_df = train_df[train_df["document"].str.len() > 100]
    train_df = train_df[train_df[label_column].notna()]

    if "Context" in label_column:
        test_set['concat_text'] = [combine_query_document(test_set.iloc[i]['Question'], test_set.iloc[i]['Document']) for i in range(len(test_set))]
    else:
        test_set['concat_text'] = [combine_query_document(test_set.iloc[i]['Question'], test_set.iloc[i]['Document'], test_set.iloc[i]['Answer']) for i in range(len(test_set))]

    train_df = train_df.drop_duplicates(["concat_text"])
    test_set = test_set.drop_duplicates(["concat_text"])

    if "Faith" in label_column:
        print("Refining data for Answer_Faithfulness classification!")
        train_df = train_df[train_df["Context_Relevance_Label"].notna()]
        train_df = train_df[train_df["Answer_Faithfulness_Label"].notna()]
        error_strings = ['answer', 'contrad', 'false', 'information', 'unanswer', 'Answer', 'Contrad', 'False', 'Information', 'Unanswer']
        train_df['generated_answer'] = train_df['generated_answer'].astype(str)
        train_df = train_df[~train_df['generated_answer'].str.contains('|'.join(error_strings))]

    return train_df, test_set

########################################################

def split_dataset(train_df, dataset, test_set, label_column):
    conversion_dict = {"Yes": 1, "No": 0}
    train_set_text = [train_df.iloc[i]['concat_text'] for i in range(len(train_df))]
    if "nq_reformatted" not in dataset:
        train_set_label = [conversion_dict[train_df.iloc[i][label_column]] for i in range(len(train_df))]
    else:
        train_set_label = [int(train_df.iloc[i][label_column]) for i in range(len(train_df))]
    
    #dev_set_text = [validation_df.iloc[i]['concat_text'] for i in range(len(validation_df))]
    #dev_set_label = [conversion_dict[validation_df.iloc[i]['Label']] for i in range(len(validation_df))]
    dev_set_text = [test_set.iloc[i]['concat_text'] for i in range(len(test_set))]
    dev_set_label = [int(test_set.iloc[i][label_column]) for i in range(len(test_set))]
    
    #test_set_text = [test_df.iloc[i]['concat_text'] for i in range(len(test_df))]
    #test_set_label = [conversion_dict[test_df.iloc[i]['Label']] for i in range(len(test_df))]
    test_set_text = [test_set.iloc[i]['concat_text'] for i in range(len(test_set))]
    test_set_label = [int(test_set.iloc[i][label_column]) for i in range(len(test_set))]

    print("Lengths of train, dev, and test")
    print(len(train_set_text))
    print(len(dev_set_text))
    print(len(test_set_text))
    print("Training example")
    print(train_set_text[100])
    print('---------------------------------------------------')
    print(train_set_label[100])
    print('---------------------------------------------------')

    labels_list = sorted(list(set(train_set_label + dev_set_label + test_set_label)))

    print("Label List")
    print(labels_list)
    print(set(train_set_label))
    print(set(dev_set_label))
    print(set(test_set_label))

    return train_set_text, train_set_label, dev_set_text, dev_set_label, test_set_text, test_set_label, labels_list

    ############################################################

def prepare_dataset(validation_set_scoring, train_set_label, train_set_text, dev_set_label, dev_set_text):
    if validation_set_scoring == True:

        training_dataset_pandas = pd.DataFrame({'label': train_set_label, 'text': train_set_text})#[:100]
        training_dataset_arrow = pa.Table.from_pandas(training_dataset_pandas)
        training_dataset_arrow = datasets.Dataset(training_dataset_arrow)

        validation_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})#[:100]
        validation_dataset_arrow = pa.Table.from_pandas(validation_dataset_pandas)
        validation_dataset_arrow = datasets.Dataset(validation_dataset_arrow)

        test_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})
        test_dataset_arrow = pa.Table.from_pandas(test_dataset_pandas)
        test_dataset_arrow = datasets.Dataset(test_dataset_arrow)

    else:

        training_dataset_pandas = pd.DataFrame({'label': train_set_label, 'text': train_set_text})#[:1000]
        training_dataset_arrow = pa.Table.from_pandas(training_dataset_pandas)
        training_dataset_arrow = datasets.Dataset(training_dataset_arrow)

        validation_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})#[:1000]
        validation_dataset_arrow = pa.Table.from_pandas(validation_dataset_pandas)
        validation_dataset_arrow = datasets.Dataset(validation_dataset_arrow)

        test_dataset_pandas = pd.DataFrame({'label': test_set_label, 'text': test_set_text})
        test_dataset_arrow = pa.Table.from_pandas(test_dataset_pandas)
        test_dataset_arrow = datasets.Dataset(test_dataset_arrow)

    return training_dataset_pandas, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow, test_dataset_pandas


def initalize_dataset_for_tokenization(tokenizer, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow):

    classification_dataset = datasets.DatasetDict({'train' : training_dataset_arrow, 
                                    'validation': validation_dataset_arrow, 
                                    'test' : test_dataset_arrow})
    
    tokenized_datasets = classification_dataset.map(lambda examples: tokenize_function(tokenizer, examples), batched=True)

    tokenized_datasets = tokenized_datasets.remove_columns(["text"])
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    tokenized_datasets.set_format("torch")

    return tokenized_datasets

    ############################################################

def train_and_evaluate_model(number_of_runs, tokenized_datasets, assigned_batch_size, train_set_label, model_choice, chosen_learning_rate, 
                            device, checkpoint_path, patience_value, num_epochs, num_warmup_steps, gradient_accumulation_multiplier): 

    micro_averages = []
    macro_averages = []
    inference_times = []

    for i in range(0, number_of_runs):

        run_start = time.time()

        print("Loading Model")

        train_dataloader = DataLoader(tokenized_datasets['train'], batch_size=assigned_batch_size)
        validation_dataloader = DataLoader(tokenized_datasets['validation'], batch_size=assigned_batch_size)
        eval_dataloader = DataLoader(tokenized_datasets['test'], batch_size=assigned_batch_size)

        print("Number of labels: " + str(len(set(train_set_label))))

        ############################################################

        model = CustomBERTModel(len(set(train_set_label)), model_choice)

        model.to(device)

        ############################################################

        criterion = nn.CrossEntropyLoss()
        optimizer = Adam(model.parameters(), lr=chosen_learning_rate)

        num_training_steps = num_epochs * len(train_dataloader)

        lr_scheduler = get_scheduler(
            name="linear", optimizer=optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_training_steps
        )

        ############################################################

        # to track the training loss as the model trains
        train_losses = []
        # to track the validation loss as the model trains
        valid_losses = []
        # to track the average training loss per epoch as the model trains
        avg_train_losses = []
        # to track the average validation loss per epoch as the model trains
        avg_valid_losses = []

        # import EarlyStopping
        from ares.LLM_as_a_Judge_Adaptation.pytorchtools import EarlyStopping

        # initialize the early_stopping object
        early_stopping = EarlyStopping(patience=patience_value, verbose=True, path=checkpoint_path)
        #early_stopping = EarlyStopping(patience=10, verbose=True)

        print("Checkpoint Path: " + checkpoint_path)

        print("Beginning Training")

        total_epochs_performed = 0

        for epoch in range(num_epochs):

            total_epochs_performed += 1

            print("Current Epoch: " + str(epoch))

            progress_bar = tqdm(range(len(train_dataloader)))

            gradient_accumulation_count = 0

            model.train()
            for batch in train_dataloader:

                if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                else:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                outputs = model(**new_batch)

                loss = criterion(outputs, batch['labels'].to(device))

                loss.backward()

                gradient_accumulation_count += 1
                if gradient_accumulation_count % (gradient_accumulation_multiplier) == 0:
                    optimizer.step()
                    lr_scheduler.step()
                    optimizer.zero_grad()
                    
                progress_bar.update(1)
                train_losses.append(loss.item())

            ######################################################

            progress_bar = tqdm(range(len(validation_dataloader)))

            model.eval()
            for batch in validation_dataloader:

                with torch.no_grad():
                
                    if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                    else:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                    if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                        new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device), 

                    outputs = model(**new_batch)

                    loss = criterion(outputs, batch['labels'].to(device))
                    progress_bar.update(1)

                    valid_losses.append(loss.item())

            # print training/validation statistics 
            # calculate average loss over an epoch
            train_loss = np.average(train_losses)
            valid_loss = np.average(valid_losses)
            avg_train_losses.append(train_loss)
            avg_valid_losses.append(valid_loss)
            
            epoch_len = len(str(num_epochs))
            
            print_msg = (f'[{epoch:>{epoch_len}}/{num_epochs:>{epoch_len}}] ' +
                        f'train_loss: {train_loss:.5f} ' +
                        f'valid_loss: {valid_loss:.5f}')
            
            print(print_msg)
            
            # clear lists to track next epoch
            train_losses = []
            valid_losses = []
            
            # early_stopping needs the validation loss to check if it has decresed, 
            # and if it has, it will make a checkpoint of the current model
            early_stopping(valid_loss, model)
            
            if early_stopping.early_stop:
                print("Early stopping")
                break
            
        return model, avg_train_losses, avg_valid_losses, eval_dataloader, inference_times

        ############################################################

def evaluate_model(model, model_choice, checkpoint_path, device, eval_dataloader, inference_times):

        print("Loading the Best Model")

        model.load_state_dict(torch.load(checkpoint_path))

        ############################################################

        print("Beginning Evaluation")

        metric = load_metric("accuracy")

        total_predictions = torch.FloatTensor([]).to(device)
        total_references = torch.FloatTensor([]).to(device)

        inference_start = time.time()

        ############################################################

        progress_bar = tqdm(range(len(eval_dataloader)))
        for batch in eval_dataloader:

            with torch.no_grad():

                if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                else:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                    new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device)

                outputs = model(**new_batch)

                logits = outputs
                predictions = torch.argmax(logits, dim=-1)
                metric.add_batch(predictions=predictions, references=batch['labels'].to(device))

                total_predictions = torch.cat((total_predictions, predictions), 0)
                total_references = torch.cat((total_references, batch['labels'].to(device)), 0)

                progress_bar.update(1)

        ############################################################

        inference_end = time.time()
        total_inference_time = inference_end - inference_start
        inference_times.append(total_inference_time)

        return total_predictions, total_references, metric

        ############################################################

def print_and_save_model(total_predictions, total_references, checkpoint_path, metric):
        print("--------------------------")
        print("Predictions and Reference Shapes")
        print(total_predictions.shape)
        print(total_references.shape)

        results = metric.compute(references=total_references, predictions=total_predictions)
        print("Accuracy for Test Set: " + str(results['accuracy']))

        f_1_metric = load_metric("f1")
        macro_f_1_results = f_1_metric.compute(average='macro', references=total_references, predictions=total_predictions)
        print("Macro F1 for Test Set: " + str(macro_f_1_results['f1'] * 100))
        micro_f_1_results = f_1_metric.compute(average='micro', references=total_references, predictions=total_predictions)
        print("Micro F1 for Test Set: " + str(micro_f_1_results['f1']  * 100))

        print("Positive / Negative Reference Ratio: " + str(round(total_references.tolist().count(1) / len(total_references.tolist()), 3)))

        print("Saved classification checkpoint to: " + str(checkpoint_path))


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument("--classification_dataset", type=str, required=True)
    parser.add_argument("--test_set_selection", type=str, required=True)
    parser.add_argument("--label_column", type=str, required=True)
    parser.add_argument("--num_epochs", type=int, required=True)
    parser.add_argument("--patience_value", type=int, required=True)
    parser.add_argument("--learning_rate", type=float, required=True)

    args = parser.parse_args()

    classification_datasets = [args.classification_dataset] # 
    test_set_selection = args.test_set_selection #
    label_column = args.label_column # 
    num_epochs = args.num_epochs # 
    patience_value = args.patience_value # 
    learning_rate_choices = [args.learning_rate] # 

    ### Instructions

    device = "cuda:0"
    device = torch.device(device)

    validation_set_scoring = True # 
    assigned_batch_size = 1
    gradient_accumulation_multiplier = 32

    number_of_runs = 1 # 
    num_warmup_steps = 100 #

    ############################################################




import os, time
import sys
#sys.path.insert(1, '../')
from ppi import clt_iid, binomial_iid, pp_mean_iid_asymptotic
import numpy as np
from tqdm import tqdm
import pandas as pd
import pdb
from sklearn.model_selection import train_test_split

######################################################################

def calculate_ppi(Y_labeled,  Yhat_labeled, Yhat_unlabeled, alpha, num_trials):
    #alpha = 0.05
    #num_trials = 500
    n_max = Y_labeled.shape[0] # Total number of labeled ballots
    ns = np.linspace(100,n_max,20).astype(int)
    #ns = np.linspace(0,n_max,20).astype(int)
    # Imputed-only estimate
    imputed_estimate = (Yhat_labeled.sum() + Yhat_unlabeled.sum())/(Yhat_labeled.shape[0] + Yhat_unlabeled.shape[0])
    # Run prediction-powered inference and classical inference for many values of n
    ci = np.zeros((num_trials, ns.shape[0], 2))
    ci_classical = np.zeros((num_trials, ns.shape[0], 2))
    for i in tqdm(range(ns.shape[0])):
        for j in range(num_trials):
            # Prediction-Powered Inference
            n = ns[i]
            rand_idx = np.random.permutation(n)
            f = Yhat_labeled.astype(float)[rand_idx[:n]]
            y = Y_labeled.astype(float)[rand_idx[:n]]    
            output = pp_mean_iid_asymptotic(y,f,Yhat_unlabeled,alpha)
            ci[j,i,:] = output
            # Classical interval
            ci_classical[j,i,:] = binomial_iid(n,alpha,y.mean())
        #print("F: " + str(f))
        #print("y: " + str(y))
        #print("Yhat_unlabeled: " + str(Yhat_unlabeled))
        #print("alpha: " + str(alpha))
        #print("output: " + str(output))  
    ci_imputed = binomial_iid(Yhat_unlabeled.shape[0], alpha, imputed_estimate)
    avg_ci = ci.mean(axis=0)[-1]
    avg_ci_classical = ci_classical.mean(axis=0)[-1]
    return avg_ci, avg_ci_classical, ci_imputed

######################################################################

column = "Context_Relevance" #"Answer_Faithfulness" #"Context_Relevance"
label_column = column + "_Label"
prediction_column = column + "_Prediction"
alpha = 0.05 #0.05
num_trials = 1000

#Y_labeled_filepath = "../datasets_v2/LLM_Judge_Test_Set_V4_Reannotated.tsv"
#Y_labeled_filepath = "../datasets_v2/LLM_Judge_Test_Set_Human_Annotations_V1.1_Filtered"
#Y_labeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/qa_logs_with_logging_details_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_with_Model_Predictions_train.tsv"
#Y_labeled_filepath = "../datasets_v2/wow_reformatted_validation_with_negatives_True_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/wow_reformatted_validation_with_negatives_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/fever_reformatted_validation_with_negatives_True_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/fever_reformatted_validation_with_negatives_False_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_True_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_False_True_False_False_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_True_True_False_False_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_False_True_False_False_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/LLM_Judge_Test_Set_Human_Annotations_V1.1_Filtered_False_True_True_True_sampled.tsv"
#Y_labeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negativesAnswer_Faithfulness_Label_with_Model_Predictions_Sampled.tsv"
Y_labeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negativesContext_Relevance_Label_with_Model_Predictions_Sampled.tsv"

Y_labeled_dataset = pd.read_csv(Y_labeled_filepath, sep="\t")
#Y_labeled_dataset[label_column] = Y_labeled_dataset["Context_Relevance"]
if 'Context_Relevance_Label_Model_Predictions' in Y_labeled_dataset.columns:
    Y_labeled_dataset["Context_Relevance_Prediction"] = Y_labeled_dataset['Context_Relevance_Label_Model_Predictions']
if 'Answer_Faithfulness_Label_Model_Predictions' in Y_labeled_dataset.columns:
    Y_labeled_dataset["Answer_Faithfulness_Prediction"] = Y_labeled_dataset['Answer_Faithfulness_Label_Model_Predictions']

Y_labeled_dataset = Y_labeled_dataset[Y_labeled_dataset[label_column].notna()]
Y_labeled = Y_labeled_dataset[label_column].values.astype(int)
Yhat_labeled = Y_labeled_dataset[prediction_column].values.astype(int)


#Yhat_unlabeled_filepath = "../../datasets/synthetic_queries_v8.1.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../../datasets/synthetic_queries_v10_with_Model_Predictions.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/qa_logs_with_logging_details_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_with_Model_Predictions_test.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/wow_reformatted_validation_with_negatives_True_True_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/wow_reformatted_validation_with_negatives_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/fever_reformatted_validation_with_negatives_True_True_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/fever_reformatted_validation_with_negatives_False_False_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_True_True_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negatives_False_True_False_False_False_True_False_False_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_True_True_False_False_True_True_False_False_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/nq_reformatted_validation_with_negatives_v2_False_True_False_False_False_True_False_False_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/qa_logs_with_logging_details_False_True_True_True_False_True_True_True_non_sampled.tsv"
#Yhat_unlabeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negativesAnswer_Faithfulness_Label_with_Model_Predictions_NonSampled.tsv"
Yhat_unlabeled_filepath = "../datasets_v2/hotpotqa_reformatted_validation_with_negativesContext_Relevance_Label_with_Model_Predictions_NonSampled.tsv"

Yhat_unlabeled_dataset = pd.read_csv(Yhat_unlabeled_filepath, sep="\t")
if 'Context_Relevance_Label_Model_Predictions' in Yhat_unlabeled_dataset.columns:
    Yhat_unlabeled_dataset["Context_Relevance_Prediction"] = Yhat_unlabeled_dataset['Context_Relevance_Label_Model_Predictions']
if 'Answer_Faithfulness_Label_Model_Predictions' in Yhat_unlabeled_dataset.columns:
    Yhat_unlabeled_dataset["Answer_Faithfulness_Prediction"] = Yhat_unlabeled_dataset['Answer_Faithfulness_Label_Model_Predictions']

Yhat_unlabeled_dataset = Yhat_unlabeled_dataset[Yhat_unlabeled_dataset[label_column].notna()]
Yhat_unlabeled_dataset = Yhat_unlabeled_dataset[Yhat_unlabeled_dataset[prediction_column].notna()]

if 'synthetic_query' in Yhat_unlabeled_dataset.columns:
    Yhat_unlabeled_dataset = Yhat_unlabeled_dataset.dropna(subset=['synthetic_query',"document", label_column])
else:
    Yhat_unlabeled_dataset = Yhat_unlabeled_dataset.dropna(subset=['Query',"Document", label_column])

######################################################################

print("Yhat_unlabeled_dataset length")
print(len(Yhat_unlabeled_dataset))

#conversion_dict = {"Yes": 1, "No": 0}
#Yhat_unlabeled_dataset[prediction_column] = Yhat_unlabeled_dataset[prediction_column].apply(lambda x: conversion_dict[x])
Yhat_unlabeled = Yhat_unlabeled_dataset[prediction_column].values.astype(int)

print(type(Y_labeled))
print(Y_labeled.shape)
print(set(Y_labeled))
print(type(Yhat_labeled))
print(Yhat_labeled.shape)
print(set(Yhat_labeled))
print(type(Yhat_unlabeled))
print(Yhat_unlabeled.shape)
print(set(Yhat_unlabeled))

print("Positive/Negative Ratio for " + label_column)
print(Yhat_unlabeled_dataset[label_column].tolist().count(1))
print(Yhat_unlabeled_dataset[label_column].tolist().count(0))
print(Yhat_unlabeled_dataset[label_column].tolist().count(1) / len(Yhat_unlabeled_dataset))
print(Yhat_unlabeled_dataset[label_column].tolist().count(0) / len(Yhat_unlabeled_dataset))

######################################################################

avg_ci, avg_ci_classical, ci_imputed = calculate_ppi(Y_labeled,  Yhat_labeled, Yhat_unlabeled, alpha, num_trials)
print("PPI Confidence Interval")
print(avg_ci)
print("Classical Confidence Interval")
print(avg_ci_classical)
print("Imputed Confidence Interval")
print(ci_imputed)


import pandas as pd
import torch
import openai
from transformers import AutoTokenizer, AutoModelForCausalLM
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring 
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_local
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_local
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_local
from tqdm.notebook import tqdm

def ues_idp_config(in_domain_prompts_dataset: str, unlabeled_evaluation_set: str, context_relevance_system_prompt: str, answer_relevance_system_prompt: str, answer_faithfulness_system_prompt: str, model_choice: str) -> dict:

    if in_domain_prompts_dataset is not None:
            in_domain_prompts_dataset = pd.read_csv(in_domain_prompts_dataset, sep='\t')
    else:
        in_domain_prompts_dataset = None 

    if unlabeled_evaluation_set is not None:
        unlabeled_evaluation_set = pd.read_csv(unlabeled_evaluation_set, sep='\t')
    else:
        unlabeled_evaluation_set = None
    if in_domain_prompts_dataset is None and unlabeled_evaluation_set is None: 
        print(f"Error: UES and IDP are not provided")
        exit()
    if in_domain_prompts_dataset is None: 
        print(f"Error: IDP is not provided")
        exit()
    if unlabeled_evaluation_set is None: 
        print(f"Error: UES is not provided")
        exit()

    context_relevance_answers = unlabeled_evaluation_set["Context_Relevance_Label"].tolist()
    answer_relevance_answers = unlabeled_evaluation_set["Answer_Relevance_Label"].tolist()
    answer_faithfulness_answers = unlabeled_evaluation_set["Answer_Faithfulness_Label"].tolist()
    context_relevance_scores = []
    answer_relevance_scores = []
    answer_faithfulness_scores = []

    gpt_models = [
    "gpt-4-0125-preview",
    "gpt-4-turbo-preview",
    "gpt-4-1106-preview",
    "gpt-4-vision-preview",
    "gpt-4-1106-vision-preview",
    "gpt-4",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-instruct",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k-0613"
]

    required_columns = ['Query', 'Document', 'Answer']
    missing_columns = [col for col in required_columns if col not in unlabeled_evaluation_set.columns]
    if missing_columns:
        print(f"Missing columns in the DataFrame: {missing_columns}")
    
    if model_choice in gpt_models:
        with tqdm(total=100, desc="Evaluating large subset with {model_choice}") as pbar:
            for index, row in unlabeled_evaluation_set[:100].iterrows():
                # Extract query, document, and answer from the row
                query = row['Query']
                document = row['Document']
                answer = row['Answer']

                # Scoring
                context_score = few_shot_context_relevance_scoring(
                    context_relevance_system_prompt, query, document, model_choice, in_domain_prompts_dataset)

                answer_relevance_score = few_shot_answer_relevance_scoring(
                    answer_relevance_system_prompt, query, document, answer, model_choice, in_domain_prompts_dataset)
                
                answer_faithfulness_score = few_shot_answer_faithfulness_scoring(answer_faithfulness_system_prompt, query, document, answer, model_choice, in_domain_prompts_dataset)

                    # Append scores to respective lists
                context_relevance_scores.append(context_score)
                answer_relevance_scores.append(answer_relevance_score)
                answer_faithfulness_scores.append(answer_faithfulness_score)

                pbar.update(1)
    else: 
        with tqdm(total=100, desc=f"Evaluating large subset with {model_choice}") as pbar:
            for index, row in unlabeled_evaluation_set[:100].iterrows():
                # Extract query, document, and answer from the row
                query = row['Query']
                document = row['Document']
                answer = row['Answer']

                # Scoring
                context_score = few_shot_context_relevance_scoring_local(
                    context_relevance_system_prompt, query, document, model_choice, in_domain_prompts_dataset)

                answer_relevance_score = few_shot_answer_relevance_scoring_local(
                    answer_relevance_system_prompt, query, document, answer, model_choice, in_domain_prompts_dataset)
                
                answer_faithfulness_score = few_shot_answer_faithfulness_scoring_local(answer_faithfulness_system_prompt, query, document, answer, model_choice, in_domain_prompts_dataset)

                # Append scores to respective lists
                context_relevance_scores.append(context_score)
                answer_relevance_scores.append(answer_relevance_score)
                answer_faithfulness_scores.append(answer_faithfulness_score)

                pbar.update(1)

    # Compile results into a dictionary
    return {
        "Context Relevance Scores": round(sum(context_relevance_scores)/len(context_relevance_scores), 3),
        "Answer Faithfulness Scores": round(sum(answer_faithfulness_scores)/len(answer_faithfulness_answers), 3),
        "Answer Relevance Scores": round(sum(answer_relevance_scores)/len(answer_relevance_answers), 3)
    }



import requests
import time
import pandas as pd
import ast
import json
import copy
import openai
from tqdm import tqdm
import csv
from datasets import Dataset
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoConfig, AutoModelForCausalLM, BitsAndBytesConfig
import torch
import numpy as np
from sklearn.model_selection import train_test_split
import re
import random
import math

#################################################################

def generate_synthetic_query_llm_approach(document: str, prompt: str, length_of_fewshot_prompt: int, device, tokenizer, model, percentiles, for_fever_dataset=False, for_wow_dataset=False):

    synthetic_queries = []

    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) +":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: "
    else:
        prompt_without_document += "Document: \nQuestion: "
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]

    if prompt_tokens_length + document_length + 100 >= 2048:
        # Added buffer for truncation
        encoded_input = tokenizer(document, max_length=2048 - prompt_tokens_length - 100, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - prompt_tokens_length - 100]) 
        document = truncated_document.replace("</s>", "")

    prompt += "Example " + str(length_of_fewshot_prompt + 1) +":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: "
    elif for_wow_dataset:
        prompt += "Dialogue: "
    else:
        prompt += "Question: "

    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    max_length = 32
    if for_wow_dataset:
        max_length = 256

    for percentile in percentiles:

        if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
            print("Length of problematic input ids: " + str(input_ids.shape))
            print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
            assert False
        outputs = model.generate(
            input_ids=input_ids,
            max_length=max_length,
            do_sample=True,
            top_p=percentile,
            num_return_sequences=1)

        query = tokenizer.decode(outputs[0], skip_special_tokens=True)
        synthetic_queries.append(query)

    return synthetic_queries

#################################################################

def generate_answer_llm_approach(document: str, question: str, prompt: str, length_of_fewshot_prompt: int, device, tokenizer, model, for_fever_dataset=False, for_wow_dataset=False):

    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) +":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: \nAnswer: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: \nResponse: "
    else:
        prompt_without_document += "Document: \nQuestion: \nAnswer: "
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]
    question_length = tokenizer.encode(question, return_tensors='pt').to(device).shape[1]

    if prompt_tokens_length + document_length + question_length + 100 >= 2048:
        # Added buffer for truncation
        reduction_length = prompt_tokens_length + question_length + 100
        encoded_input = tokenizer(document, max_length=2048 - reduction_length, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - reduction_length]) 
        document = truncated_document.replace("</s>", "")

    prompt += "Example " + str(length_of_fewshot_prompt + 1) +":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Answer: " 
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Response: " 
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Answer: " 

    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
        print("Length of problematic input ids: " + str(input_ids.shape))
        print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
        assert False
    outputs = model.generate(
        input_ids=input_ids,
        max_length=256,
        do_sample=True,
        top_p=0.05,
        num_return_sequences=1)

    query = tokenizer.decode(outputs[0], skip_special_tokens=True)

    return query

#################################################################

def generate_synthetic_query_openai_approach(document: str, system_prompt: str, few_shot_examples: str, temperatures: list, length_of_fewshot_prompt: int):

    time.sleep(1)

    synth_documents_generated = []
    #try:
    for temp in temperatures:
        try:
            user_prompt = few_shot_examples
            #user_prompt += "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Question: "

            messages = [
                {
                    "role": "system",
                    "content": system_prompt
                },
                {
                    "role": "user",
                    "content": user_prompt
                }
            ]

            response = openai.ChatCompletion.create(
                    model="gpt-3.5-turbo-16k",
                    #model="gpt-4",
                    messages=messages,
                    temperature=temp
                )
                
            final_response = response["choices"][0]["message"]["content"]
            synth_documents_generated.append(final_response)
        except Exception as E:
            print("Error with OpenAI! Waiting one minute...")
            print("Error: " + str(E))
            time.sleep(30)
        #return final_response
    #except:
    #    print("Error Querying OpenAI! Attempting again...")

    assert len(synth_documents_generated) >= 1
    return synth_documents_generated

#################################################

def generate_answer_from_context(document: str, synth_question: str):

    time.sleep(1)

    for _ in range(5):

        try:

            system_prompt = "You are a helpful assistant built by Databricks, you are not human, you are good at helping to answer a query based on the context step by step, the context is a document. If the query doesn't form a complete question, just say I don't know. If there is a good answer from the context, try to summarize the context as the answer. If you don't know the answer, just say I don't know. If there is no enough information to determine the answer, just say I don't know. If the context is irrelevant to the question, just say I don't know."
            user_prompt = f"Here is the question {synth_question}\nHere is the context: {(document)}?"

            messages = [
                    {
                        "role": "system",
                        "content": system_prompt
                    },
                    {
                        "role": "user",
                        "content": user_prompt
                    }
                ]

            response = openai.ChatCompletion.create(
                        model="gpt-3.5-turbo-16k",
                        messages=messages,
                        temperature=0.0
                    )
                    
            final_response = response["choices"][0]["message"]["content"]
            return final_response

        except:

            print("Error querying OpenAI! Attempting again...")

#################################################

def generate_contradictory_answer_from_context(document: str, synth_question: str):

    time.sleep(1)

    for _ in range(5):

        try:

            #system_prompt = "You are a helpful assistant built by Databricks, you are not human, you are good at helping to answer a query based on the context step by step, the context is a document. If the query doesn't form a complete question, just say I don't know. If there is a good answer from the context, try to summarize the context as the answer. If you don't know the answer, just say I don't know. If there is no enough information to determine the answer, just say I don't know. If the context is irrelevant to the question, just say I don't know."
            system_prompt = "Create an answer for the given question that contradicts the provided document. You should create false information that disagrees with what exists within the content of the document."
            user_prompt = f"Question: {synth_question}\nDocument:{(document)}"
            user_prompt = system_prompt + "\n\n" + user_prompt

            messages = [
                    #{
                    #    "role": "system",
                    #    "content": system_prompt
                    #},
                    {
                        "role": "user",
                        "content": user_prompt
                    }
                ]

            response = openai.ChatCompletion.create(
                        model="gpt-3.5-turbo-16k",
                        messages=messages
                    )
                    
            final_response = response["choices"][0]["message"]["content"]
            return final_response

        except:

            print("Error querying OpenAI! Attempting again...")

#################################################

def check_generated_answer(answer: str):
    problematic_phrases = ["I don't know", "don't know", "i don't know"]
    for phrase in problematic_phrases:
        if phrase in answer:
            return "No"
    return "Yes"

def generate_contradictory_answer_examples(queries_dataset, number_of_contradictory_answers_to_generate: int, few_shot_examples_for_contradictory_answers=None, 
                                           device=None, tokenizer=None, model=None, for_fever_dataset=None, for_wow_dataset=None):

    def remove_problematic_contradictory_phrases(text):
        if text is None:
            return text

        problematic_phrases = ["Contradictory Answer:", "The false information created is:", "Incorrect Answer: "]
        text_split = text.split(":")
        if len(text_split) > 1:
            return text_split[1]
        else:
            return text

    queries_dataset_copy = queries_dataset.copy()
    #queries_dataset_copy = queries_dataset_copy[queries_dataset_copy['Context_Relevance_Label'] == "Yes"]
    queries_dataset_copy = queries_dataset_copy.drop_duplicates(subset=['synthetic_query'])

    number_of_contradictory_answers_to_generate = min(number_of_contradictory_answers_to_generate, len(queries_dataset_copy))
    queries_dataset_copy = queries_dataset_copy.sample(n=number_of_contradictory_answers_to_generate, random_state=42)

    contradictory_answers = []
    contradictory_labels = []
    for i in tqdm(range(len(queries_dataset_copy))):

        if few_shot_examples_for_contradictory_answers is None:
            contradictory_answer_generated = generate_contradictory_answer_from_context(queries_dataset_copy.iloc[i]['document'], queries_dataset_copy.iloc[i]['synthetic_query'])
        else:
            contradictory_answer_generated = generate_contradictory_answer_llm_approach(queries_dataset_copy.iloc[i]['document'], queries_dataset_copy.iloc[i]['synthetic_query'], few_shot_examples_for_contradictory_answers, device, tokenizer, model, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset)

        contradictory_answer_generated = remove_problematic_contradictory_phrases(contradictory_answer_generated)

        contradictory_answers.append(contradictory_answer_generated)   
        contradictory_labels.append("No")

    queries_dataset_copy['generated_answer'] = contradictory_answers
    queries_dataset_copy['Answer_Faithfulness_Label'] = contradictory_labels
    queries_dataset_copy['generated_answer'] = contradictory_answers
    queries_dataset_copy['Answer_Relevance_Label'] = contradictory_labels

    print("Contradictory Answers Added using Contradiction Generation")
    print(len(queries_dataset_copy))

    #################################################

    queries_dataset_copy_2 = queries_dataset.copy()
    queries_dataset_copy_2 = queries_dataset_copy_2.drop_duplicates(subset=['synthetic_query'])

    queries_dataset_copy_2 = queries_dataset_copy_2.sample(n=number_of_contradictory_answers_to_generate, random_state=42)
    total_answers = queries_dataset[queries_dataset['Answer_Relevance_Label'] == "Yes"]['generated_answer'].tolist()
    total_answers = [answer for answer in total_answers if isinstance(answer, str)]
    total_answers = [str(answer) for answer in total_answers]
    total_answers = [answer for answer in total_answers if len(answer) > 5]

    contradictory_answers_2 = []
    contradictory_labels_2 = []
    for i in tqdm(range(len(queries_dataset_copy_2))):

        random_answer = random.choice(total_answers)
        contradictory_answers_2.append(random_answer)   
        contradictory_labels_2.append("No")

    queries_dataset_copy_2['generated_answer'] = contradictory_answers_2
    queries_dataset_copy_2['Answer_Relevance_Label'] = contradictory_labels_2

    print("Contradictory Answers Added using Answer Randomization")
    print(len(queries_dataset_copy_2))

    #################################################

    # Shuffle dataframe
    queries_dataset = pd.concat([queries_dataset, queries_dataset_copy, queries_dataset_copy_2], axis=0, ignore_index=True)
    queries_dataset = queries_dataset.sample(n=len(queries_dataset), random_state=42)

    # contradictory_answers_df = queries_dataset[queries_dataset['Answer_Relevance_Label'] == "No"]

    # # Print each contradictory answer with its query
    # for index, row in contradictory_answers_df.iterrows():
    #     print(f"Query: {row['synthetic_query']}, Contradictory Answer: {row['generated_answer']}")

    # breakpoint()

    return queries_dataset

#################################################

def generate_contradictory_answer_llm_approach(document: str, question: str, prompt: str, device, tokenizer, model, for_fever_dataset=False, for_wow_dataset=False):

    prompt_without_document = prompt + "Example " + str(prompt.count("Example") + 1) +":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: \nIncorrect Answer: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: \nIncorrect Response: "
    else:
        prompt_without_document += "Document: \nQuestion: \nIncorrect Answer: "
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]
    question_length = tokenizer.encode(question, return_tensors='pt').to(device).shape[1]

    if prompt_tokens_length + document_length + question_length + 100 >= 2048:
        # Added buffer for truncation
        reduction_length = prompt_tokens_length + question_length + 100
        encoded_input = tokenizer(document, max_length=2048 - reduction_length, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - reduction_length]) 
        document = truncated_document.replace("</s>", "")

    prompt += "Example " + str(prompt.count("Example") + 1) +":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Incorrect Answer: " 
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Incorrect Response: " 
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Incorrect Answer: " 

    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    #print("New input ids")
    #print(input_ids.shape)
    #assert False

    if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
        print("Length of problematic input ids: " + str(input_ids.shape))
        print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
        assert False
    outputs = model.generate(
        input_ids=input_ids,
        max_length=256,
        do_sample=True,
        top_p=1.0,
        num_return_sequences=1)

    query = tokenizer.decode(outputs[0], skip_special_tokens=True)

    return query




import numpy as np
import matplotlib.pyplot as plt
import itertools
from scipy.stats import binom, norm
from scipy.special import expit
from scipy.optimize import brentq
from sklearn.linear_model import LogisticRegression
from joblib import delayed, Parallel
import pdb

"""
    IID Concentration Bounds
"""
def binomial_iid(N,alpha,muhat):
    def invert_upper_tail(mu): return binom.cdf(N*muhat, N, mu) - (alpha/2)
    def invert_lower_tail(mu): return binom.cdf(N*muhat, N, mu) - (1-alpha/2)
    u = brentq(invert_upper_tail,0,1)
    l = brentq(invert_lower_tail,0,1)
    return np.array([l,u])

def bentkus_iid(N, alpha, muhat):
    return binomial_iid(N, alpha/np.e, muhat)

def clt_iid(x, alpha):
    n = x.shape[0]
    sigmahat = x.std()
    w = norm.ppf(1-alpha/2) * sigmahat / np.sqrt(n)
    return np.array([ x.mean() - w, x.mean() + w ])

def wsr_iid(x_n, alpha, grid, num_cpus=10, parallelize: bool = False, intersection: bool = True,
            theta: float = 0.5, c: float = 0.75):
    n = x_n.shape[0]
    t_n = np.arange(1, n + 1)
    muhat_n = (0.5 + np.cumsum(x_n)) / (1 + t_n)
    sigma2hat_n = (0.25 + np.cumsum(np.power(x_n - muhat_n, 2))) / (1 + t_n)
    sigma2hat_tminus1_n = np.append(0.25, sigma2hat_n[: -1])
    assert(np.all(sigma2hat_tminus1_n > 0))
    lambda_n = np.sqrt(2 * np.log(2 / alpha) / (n * sigma2hat_tminus1_n))

    def M(m):
        lambdaplus_n = np.minimum(lambda_n, c / m)
        lambdaminus_n = np.minimum(lambda_n, c / (1 - m))
        return np.maximum(
            theta * np.exp(np.cumsum(np.log(1 + lambdaplus_n * (x_n - m)))),
            (1 - theta) * np.exp(np.cumsum(np.log(1 - lambdaminus_n * (x_n - m))))
        )

    if parallelize:  # sometimes much slower
        M = np.vectorize(M)
        M_list = Parallel(n_jobs=num_cpus)(delayed(M)(m) for m in grid)
        indicators_gxn = np.array(M_list) < 1 / alpha
    else:
        indicators_gxn = np.zeros([grid.size, n])
        found_lb = False
        for m_idx, m in enumerate(grid):
            m_n = M(m)
            indicators_gxn[m_idx] = m_n < 1 / alpha
            if not found_lb and np.prod(indicators_gxn[m_idx]):
                found_lb = True
            if found_lb and not np.prod(indicators_gxn[m_idx]):
                break  # since interval, once find a value that fails, stop searching
    if intersection:
        ci_full = grid[np.where(np.prod(indicators_gxn, axis=1))[0]]
    else:
        ci_full =  grid[np.where(indicators_gxn[:, -1])[0]]
    if ci_full.size == 0:  # grid maybe too coarse
        idx = np.argmax(np.sum(indicators_gxn, axis=1))
        if idx == 0:
            return np.array([grid[0], grid[1]])
        return np.array([grid[idx - 1], grid[idx]])
    return np.array([ci_full.min(), ci_full.max()]) # only output the interval

"""
    Mean estimation confidence intervals
"""

def pp_mean_iid_asymptotic(Y_labeled, Yhat_labeled, Yhat_unlabeled, alpha):
    n = Y_labeled.shape[0]
    N = Yhat_unlabeled.shape[0]
    tildethetaf = Yhat_unlabeled.mean()
    rechat = (Yhat_labeled - Y_labeled).mean()
    thetahatPP = tildethetaf - rechat
    sigmaftilde = np.std(Yhat_unlabeled)
    sigmarec = np.std(Yhat_labeled -  Y_labeled)
    hw = norm.ppf(1-alpha/2)*np.sqrt((sigmaftilde**2/N) + (sigmarec**2/n))
    return [thetahatPP - hw, thetahatPP + hw]

"""
    OLS algorithm with sandwich variance estimator
"""
def ols(features, outcome):
    ols_coeffs = np.linalg.pinv(features).dot(outcome)
    return ols_coeffs

def classical_ols_interval(X, Y, alpha, return_stderr=False, sandwich=True):
    n = X.shape[0]
    thetahat = ols(X, Y)
    Sigmainv = np.linalg.inv(1/n * X.T@X)
    if sandwich:
        M = 1/n * (X.T*((Y - X@thetahat)**2)[None,:])@X
    else:
        M = 1/n * ((Y - X@thetahat)**2).mean() * X.T@X
    V = Sigmainv@M@Sigmainv
    stderr = np.sqrt(np.diag(V))
    if return_stderr:
        return stderr
    halfwidth = norm.ppf(1-alpha/2) * stderr/np.sqrt(n)
    return thetahat - halfwidth, thetahat + halfwidth

def pp_ols_interval(X_labeled, X_unlabeled, Y_labeled, Yhat_labeled, Yhat_unlabeled, alpha, sandwich=True):
    n = X_labeled.shape[0]
    N = X_unlabeled.shape[0]
    thetatildef = ols(X_unlabeled, Yhat_unlabeled)
    rectifierhat = ols(X_labeled, Y_labeled - Yhat_labeled)
    pp_thetahat = thetatildef + rectifierhat
    stderr_tildef = classical_ols_interval(X_unlabeled, Yhat_unlabeled, 0.001*alpha, return_stderr=True, sandwich=sandwich)
    stderr_rec = classical_ols_interval(X_labeled, Y_labeled-Yhat_labeled, 0.999*alpha, return_stderr=True, sandwich=sandwich)
    halfwidth = norm.ppf(1-alpha/2)*np.sqrt(((stderr_rec**2)/n) + ((stderr_tildef**2)/N))
    return pp_thetahat - halfwidth, pp_thetahat + halfwidth

"""
    Logistic regression algorithm
"""
def logistic(X, y):
    clf = LogisticRegression(penalty='none', solver='lbfgs', max_iter=10000, tol=1e-15, fit_intercept=False).fit(X,y)
    return clf.coef_.squeeze()

def product(*args, **kwds):
    # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
    # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
    pools = map(tuple, args)
    result = [[]]
    for pool in pools:
        result = [x+[y] for x in result for y in pool]
    for prod in result:
        yield tuple(prod)

def classical_logistic_interval(X, Y, alpha, num_grid=500):
    n = X.shape[0]
    d = X.shape[1]
    Y = (Y >= 0.5).astype(int)
    point_estimate = logistic(X, Y) # Used for setting the grid.

    theta_grid = np.concatenate([
        np.linspace(-3*point_estimate, point_estimate, num_grid//2),
        np.linspace(point_estimate, 3*point_estimate, num_grid//2)[1:]
    ])

    mu = expit(X@theta_grid.T)
    g = 1/n * X.T@(mu - Y[:, None])

    sigmahat_err = np.std(X[:,:,None]*(mu - Y[:,None])[:,None,:], axis=0)
    grad_halfwidth = norm.ppf(1-alpha/(2*d)) * sigmahat_err/np.sqrt(n)

    condition = np.all(np.abs(g) <= grad_halfwidth, axis=0)

    Cpp = theta_grid[condition]

    # TODO: If all positive, make grid wider
    assert (condition[0] == False) & (condition[-1] == False)

    return [ Cpp.min(axis=0), Cpp.max(axis=0) ]

def pp_logistic_interval(X_labeled, X_unlabeled, Y_labeled, Yhat_labeled, Yhat_unlabeled, alpha, num_grid=500):
    X = np.concatenate([X_labeled, X_unlabeled], axis=0)
    n = X_labeled.shape[0]
    d = X_labeled.shape[1]
    N = X_unlabeled.shape[0]
    Yhat_labeled = np.clip(Yhat_labeled, 0, 1)
    Yhat_unlabeled = np.clip(Yhat_unlabeled, 0, 1)

    Yhat = np.concatenate([Yhat_labeled, Yhat_unlabeled], axis=0)
    point_estimate = logistic(X_labeled, (Y_labeled > 0.5).astype(int))

    rechat = 1/n * X_labeled.T @ (Yhat_labeled - Y_labeled)
    sigmahat_rec = np.std(X_labeled * (Yhat_labeled - Y_labeled)[:,None], axis=0)

    theta_grid = np.concatenate([
        np.linspace(-3*point_estimate, point_estimate, num_grid//2),
        np.linspace(point_estimate, 3*point_estimate, num_grid//2)[1:]
    ])

    mu = expit(X_unlabeled@theta_grid.T)
    g = 1/N * X_unlabeled.T@(mu - Yhat_unlabeled[:, None])

    sigmahat_err = np.std(X_unlabeled[:,:,None]*(mu - Yhat_unlabeled[:,None])[:,None,:], axis=0)

    halfwidth = norm.ppf(1-alpha/(2*d)) * np.sqrt( sigmahat_rec[:,None]**2/n + sigmahat_err**2/N)

    condition = np.all( np.abs(g + rechat[:,None]) <= halfwidth, axis=0)

    Cpp = theta_grid[condition]

    # TODO: If all positive, make grid wider
    assert (condition[0] == False) & (condition[-1] == False)

    return [ Cpp.min(axis=0), Cpp.max(axis=0) ]

"""
    DISCRETE L_p ESTIMATION RATES
"""

def linfty_dkw(N, K, alpha):
    return np.sqrt(2/N * np.log(2 / alpha))

def linfty_binom(N, K, alpha, qhat):
    epsilon = 0
    for k in np.arange(K):
        bci = binomial_iid(N, alpha/K, qhat[k])
        epsilon = np.maximum(epsilon, np.abs(bci-qhat[k]).max())
    return epsilon

"""
	SAMPLING WITHOUT REPLACEMENT
"""
def clt_swr(x,N,alpha):
    n = x.shape[0]
    point_estimate = x.mean()
    fluctuations = x.std()*norm.cdf(1-alpha/2)*np.sqrt((N-n)/(N*n))
    return np.array([point_estimate-fluctuations, point_estimate+fluctuations])

def wsr_swr(x,N,alpha,grid,num_cpus=10, intersection=True): # x is a [0,1] bounded sequence
    n = x.shape[0]
    def mu(m,i): return (N*m - np.concatenate([np.array([0,]), np.cumsum(x[:i-1])]))/(N - (np.arange(i)+1) + 1 )
    muhats = (1/2 + np.cumsum(x))/(np.arange(n)+1)
    sigmahat2s = (1/4 + np.cumsum((x-muhats)**2))/(np.arange(n)+1)
    lambdas = np.concatenate([np.array([1,]), np.sqrt(2*np.log(2/alpha)/(n*sigmahat2s))[:-1]]) # can't use last entry
    def M(m,i): return 1/2*np.maximum(
        np.prod(1+np.minimum(lambdas[:i], 0.5/mu(m,i))*(x[:i]-mu(m,i))),
        np.prod(1-np.minimum(lambdas[:i], 0.5/(1-mu(m,i)))*(x[:i]-mu(m,i)))
    )
    M = np.vectorize(M)
    if intersection:
        M_list = Parallel(n_jobs=num_cpus)(delayed(M)(grid,i) for i in range(1,n+1))
    else:
        M_list =[M(grid, n),]
    ci_full = grid[np.where(np.prod(np.stack(M_list, axis=1) < 1/alpha , axis=1))[0]]
    return np.array([ci_full.min(), ci_full.max()]) # only output the interval


from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import begin
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import filter_dataset
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import preprocess_data
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import load_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import post_process_predictions
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_and_scoring_data

def rag_scoring_config(alpha, num_trials, evaluation_datasets, few_shot_examples_filepath, checkpoints, labels,
GPT_scoring, swap_human_labels_for_gpt4_labels, model_choice, assigned_batch_size, number_of_labels, 
use_pseudo_human_labels, gold_label_path):

    for checkpoint, label_column in zip(checkpoints, labels):
        LLM_judge_ratio_predictions = []
        validation_set_lengths = []
        validation_set_ratios = []
        ppi_confidence_intervals = []
        accuracy_scores = []
        for test_set_selection in evaluation_datasets:

            few_shot_examples = begin(evaluation_datasets, checkpoints, labels, GPT_scoring, few_shot_examples_filepath)

            context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt = filter_dataset(evaluation_datasets)
            
            test_set, text_column = preprocess_data(test_set_selection, label_column, labels)

            model, tokenizer, device = load_model(model_choice, number_of_labels, GPT_scoring, checkpoint) 

            total_predictions, total_references, results, metric = evaluate_model(test_set, label_column, text_column, device, GPT_scoring, tokenizer, model, assigned_batch_size, model_choice, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt)

            test_set, Y_labeled_dataset, Y_labeled_dataloader, Y_labeled_predictions, Yhat_unlabeled_dataset, prediction_column = post_process_predictions(test_set, label_column, total_predictions, labels, use_pseudo_human_labels, gold_label_path, tokenizer, assigned_batch_size, device) 

            evaluate_and_scoring_data(test_set, Y_labeled_predictions, Y_labeled_dataset, Y_labeled_dataloader, Yhat_unlabeled_dataset, alpha, num_trials, model, device, model_choice, 
            swap_human_labels_for_gpt4_labels, context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt,
            few_shot_examples, metric, prediction_column, label_column, test_set_selection, 
            LLM_judge_ratio_predictions, validation_set_lengths, validation_set_ratios, 
            ppi_confidence_intervals, accuracy_scores, results)



import requests
import time
import pandas as pd
import ast
import json
import copy
import openai
from openai import OpenAI
from tqdm import tqdm
import csv
from datasets import Dataset
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoConfig, AutoModelForCausalLM
import torch
import numpy as np
from sklearn.model_selection import train_test_split
import requests 
import os
import re
import time

####################################################################

def calculate_accuracy(predictions, ground_truth):
    if len(predictions) != len(ground_truth):
        print(len(predictions))
        print(len(ground_truth))
        raise ValueError("Input lists must have the same length")

    correct_count = sum(1 for pred, truth in zip(predictions, ground_truth) if pred == truth)
    total_count = len(predictions)

    accuracy = round(correct_count * 100 / total_count, 2)
    return accuracy

####################################################################

def few_shot_context_relevance_scoring(system_prompt: str, query: str, document: str, gpt_model: str, few_shot_examples=None):

    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Label: "
            
            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content


            if "[[Yes]]" in final_response:
                return 1
            elif "[[No]]" in final_response:
                return 0
            else:
                print("Didn't extract Yes or No!")
                return 1
        #except:
        #    print("Error with querying OpenAI! Trying again...")
        #    time.sleep(60)
    
####################################################################

def few_shot_answer_faithfulness_scoring(system_prompt, query: str, document: str, answer: str, gpt_model: str, few_shot_examples=None):

    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Answer: " + answer + "\n"
            user_prompt += "Label: "
            
            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if "[[Yes]]" in final_response:
                return 1
            elif "[[No]]" in final_response:
                return 0
            else:
                print("Didn't extract Yes or No!")
                return 1
        #except:
            print("Error with querying OpenAI! Trying again...")
            time.sleep(60)
    
####################################################################

def few_shot_answer_relevance_scoring(system_prompt: str, query: str, document: str, answer: str, gpt_model: str, few_shot_examples=None):

    for _ in range(5):
        #try:

            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                    user_prompt += "Label: " + str(current_label) + "\n\n"

            user_prompt += "Question: " + query + "\n"
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Answer: " + answer + "\n"
            user_prompt += "Label: "
            
            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content


            if "[[Yes]]" in final_response:
                return 1
            elif "[[No]]" in final_response:
                return 0
            else:
                print("Didn't extract Yes or No!")
                return 1
        #except:
            print("Error with querying OpenAI! Trying again...")
            time.sleep(60)


##############################CUSTOM MODELS (NOT OPENAI)######################################
            
def few_shot_context_relevance_scoring_local(system_prompt: str, query: str, document: str, model_choice, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        print("TOGETHER_API_KEY environment variable not set. Please set the variable.")
        return -1

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                    user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
                    user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Label: "

    time.sleep(4)

    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content

    if "[[Yes]]" or "Yes" in final_response:
        print(final_response)
        return 1
    elif "[[No]]" or "No" in final_response:
        print(final_response)
        return 0
    else:
        breakpoint()
        print("Didn't extract Yes or No!")
        print(f"Unexpected response received: {final_response}")
        return 1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1

def few_shot_answer_faithfulness_scoring_local(system_prompt: str, query: str, document: str, answer: str, model_choice, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        print("TOGETHER_API_KEY environment variable not set. Please set the variable.")
        return -1

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
                user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
                user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
                user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
                current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + answer + "\n"
    user_prompt += "Label: "

    time.sleep(4)

    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content

    if "[[Yes]]" or "Yes" in final_response:
        print(final_response)
        return 1
    elif "[[No]]" or "No" in final_response:
        print(final_response)
        return 0
    else:
        breakpoint()
        print("Didn't extract Yes or No!")
        print(f"Unexpected response received: {final_response}")
        return 1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1

def few_shot_answer_relevance_scoring_local(system_prompt: str, query: str, document: str, answer: str, model_choice, few_shot_examples=None):
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    if not TOGETHER_API_KEY:
        print("TOGETHER_API_KEY environment variable not set. Please set the variable.")
        return -1

    user_prompt = ""
    if few_shot_examples is not None:
        for row in range(len(few_shot_examples)):
            user_prompt += "Question: " + few_shot_examples.iloc[row]['Query'] + "\n"
            user_prompt += "Document: " + few_shot_examples.iloc[row]['Document'] + "\n"
            user_prompt += "Answer: " + few_shot_examples.iloc[row]['Answer'] + "\n"
            current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
            user_prompt += "Label: " + str(current_label) + "\n\n"

    user_prompt += "Question: " + query + "\n"
    user_prompt += "Document: " + document + "\n"
    user_prompt += "Answer: " + answer + "\n"
    user_prompt += "Label: "

    time.sleep(4)

    client = OpenAI(
    api_key=TOGETHER_API_KEY,
    base_url='https://api.together.xyz/v1',
    )

    chat_completion = client.chat.completions.create(
    messages=[
        {
        "role": "system",
        "content": system_prompt,
        },
        {
        "role": "user",
        "content": user_prompt,
        }
    ],
    model=model_choice
    )
    final_response = chat_completion.choices[0].message.content

    if "[[Yes]]" or "Yes" in final_response:
        print(final_response)
        return 1
    elif "[[No]]" or "No" in final_response:
        print(final_response)
        return 0
    else:
        print("Didn't extract Yes or No!")
        print(f"Unexpected response received: {final_response}")
        return 1
    # except requests.exceptions.RequestException as e:
    #     print(f"HTTP Request failed: {e}")
    #     return -1


from ares.ares import ARES
import pandas as pd


# ues_idp_config = {
#     # Dataset for in-domain prompts
#     "in_domain_prompts_dataset": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
    
#     # Dataset for unlabeled evaluation
#     "unlabeled_evaluation_set": "/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.5.tsv", 

#     "context_relevance_system_prompt": """You are an expert dialogue agent. Your task is to analyze the provided document and determine whether 
#     it is relevant for responding to the dialogue. Consider the content of the document and its relation to the provided dialogue. 
#     Output your final verdict in the format: "[[Yes]]" if the document is relevant, and "[[No]]" if the document provided is not relevant. 
#     Strictly adhere to this response format without any additional explanation.
#     """,

#     "answer_relevance_system_prompt": """Given a question, a document, and an answer, analyze both the provided answer and document to determine if the answer is relevant to the question. 
#     Evaluate whether the answer addresses all aspects of the question and relies solely on correct information from the document for its response. 
#     Output your final verdict in the format: "[[Yes]]" if the answer is relevant to the given question, and "[[No]]" if the answer is not relevant. 
#     Maintain strict adherence to this response format without additional explanation. """,
    
#     "answer_faithfulness_system_prompt": """Given a question, a document, and an answer, assess whether the provided answer is faithful to the document's contents. 
#     The answer must neither introduce information beyond what is contained in the document nor contradict any document information. 
#     Output your verdict in the format: "[[Yes]]" if the answer is faithful to the document, and "[[No]]" if the answer is unfaithful. 
#     Ensure strict adherence to this response format without providing additional details.
#     """,

#     "model_choice" : "meta-llama/Llama-2-70b-chat-hf"
# }

# ues_idp_config = {
#     # Dataset for in-domain prompts
#     "in_domain_prompts_dataset": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
    
#     # Dataset for unlabeled evaluation
#     "unlabeled_evaluation_set": "/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.5.tsv", 
    
#     # Model: Mistral 7B
#     "model_choice" : "gpt-3.5-turbo-0125"
# }

# synth_config = { 
#     "document_filepaths": ["/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.5.tsv"],
#     "few_shot_prompt_filename": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "synthetic_queries_filenames": ["data/output/synthetic_queries_1.tsv"],
#     "documents_sampled": 20
# }

# # HOTPOTQA
# synth_config = { 
#     "document_filepaths": ["/future/u/manihani/ARES/data/datasets_v2/hotpotqa/ratio_0.6_reformatted_full_articles_False_validation_with_negatives.tsv"],
#     "few_shot_prompt_filename": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "synthetic_queries_filenames": ["data/output/hotpotqa_synthetc_queries_1.tsv"],
#     "documents_sampled": 50
# }

# ppi_config = { 
#     "evaluation_datasets": ['datasets_v2/nq/ratio_0.525_reformatted_full_articles_False_validation_with_negatives.tsv'], 
#     "few_shot_examples_filepath": "datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "checkpoints": ["/future/u/manihani/ARES/example_checkpoints/Context-Relevance-Label.pt", "/future/u/manihani/ARES/example_checkpoints/Answer-Relevance-Validation-1867825.pt"],
#     "labels": ["Context_Relevance_Label", "Answer_Relevance_Label"], 
#     "GPT_scoring": False, 
#     "gold_label_path": "datasets_v2/nq/ratio_0.525_reformatted_full_articles_False_validation_with_negatives.tsv", 
#     "swap_human_labels_for_gpt4_labels": False
# }

# WoW 
# synth_config = { 
#     "document_filepaths": ["/future/u/manihani/ARES/data/datasets_v2/wow/ratio_0.6_reformatted_full_articles_False_validation_with_negatives.tsv"],
#     "few_shot_prompt_filename": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "synthetic_queries_filenames": ["data/output/WoW_synthetic_queries_1.tsv"],
#     "documents_sampled": 10000
# }


# classifier_config = {
#     "classification_dataset": "output/synthetic_queries_1.tsv", 
#     "test_set_selection": "/future/u/manihani/ARES/datasets_v2/nq/ratio_0.6_reformatted_full_articles_False_validation_with_negatives.tsv", 
#     "label_column": "Answer_Relevance_Label", 
#     "num_epochs": 10, 
#     "patience_value": 3, 
#     "learning_rate": 5e-6
# }

# TEST 
classifier_config = {
    "classification_dataset": "data/output/synthetic_queries_1.tsv", 
    "validation_set": "/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.6.tsv", 
    "label_column": "Answer_Relevance_Label", 
    "num_epochs": 10, 
    "patience_value": 3, 
    "learning_rate": 5e-6
}

# ppi_config = { 
#     "evaluation_datasets": ['/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.6.tsv'], 
#     "few_shot_examples_filepath": "/future/u/manihani/ARES/data/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "checkpoints": ["/future/u/manihani/ARES/data/example_checkpoints/5e-06_1_True_Context_Relevance_Label_ratio_0.6_reformatted_full_articles_False_validation_with_negatives_428380.pt"],
#     "labels": ["Context_Relevance_Label"], 
#     "GPT_scoring": False, 
#     "gold_label_path": "/future/u/manihani/ARES/data/datasets_v2/nq/nq_ratio_0.525.tsv", 
#     "swap_human_labels_for_gpt4_labels": False
# }

# ppi_config = { 
#     "evaluation_datasets": ['datasets_v2/nq/ratio_0.525_reformatted_full_articles_False_validation_with_negatives.tsv'], 
#     "few_shot_examples_filepath": "datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "checkpoints": ["/future/u/manihani/ARES/example_checkpoints/Context-Relevance-Label.pt", "/future/u/manihani/ARES/example_checkpoints/Answer-Relevance-Validation-1867825.pt"],
#     "labels": ["Context_Relevance_Label", "Answer_Relevance_Label"], 
#     "GPT_scoring": False, 
#     "gold_label_path": "datasets_v2/nq/ratio_0.525_reformatted_full_articles_False_validation_with_negatives.tsv", 
#     "swap_human_labels_for_gpt4_labels": False
# }

# ppi_config = { 
#     "evaluation_datasets": ['/future/u/manihani/ARES/datasets_v2/nq/ratio_0.5_reformatted_full_articles_False_validation_with_negatives.tsv'], 
#     "few_shot_examples_filepath": "/future/u/manihani/ARES/datasets/multirc_few_shot_prompt_for_synthetic_query_generation_v1.tsv",
#     "checkpoints": ["/future/u/manihani/ARES/checkpoints/microsoft-deberta-v3-large/output-synthetic_queries_1.tsv/5e-06_1_True_Context_Relevance_Label_ratio_0.6_reformatted_full_articles_False_validation_with_negatives_428380.pt"],
#     "labels": ["Context_Relevance_Label"], 
#     "GPT_scoring": False, 
#     "gold_label_path": "/future/u/manihani/ARES/datasets_v2/nq/ratio_0.5_reformatted_full_articles_False_validation_with_negatives.tsv", 
#     "swap_human_labels_for_gpt4_labels": False
# }

# ares = ARES(ues_idp=ues_idp_config)
# results = ares.ues_idp()
# print(results)

# ares = ARES(synthetic_query_generator=synth_config)
# results = ares.generate_synthetic_data()
# print(results)

ares = ARES(classifier_model=classifier_config)
results = ares.train_classifier()
print(results)

# ares = ARES(ppi=ppi_config)
# results = ares.evaluate_RAG()
# print(results)

# ares = ARES() 
# ares.KILT_dataset("fever")


from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import load_model
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import prepare_and_clean_data
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import analyze_and_report_data
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import transform_data
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import split_dataset
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import prepare_dataset
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import initalize_dataset_for_tokenization
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import train_and_evaluate_model
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import evaluate_model 
from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import print_and_save_model
import torch

def binary_classifer_config(classification_dataset: str, validation_set: str, label_column: str, 
                    num_epochs: int, patience_value: int, learning_rate: float, 
                    model_choice: str = "microsoft/deberta-v3-large", validation_set_scoring: bool = True, 
                    assigned_batch_size: int = 1, gradient_accumulation_multiplier: int = 32, 
                    number_of_runs: int = 1, num_warmup_steps: int = 100):

    device = "cuda:0"
    device = torch.device(device)
    learning_rate_choices = [learning_rate]
    classification_datasets = [classification_dataset]

    for chosen_learning_rate in learning_rate_choices:
        for dataset in classification_datasets:

            tokenizer, max_token_length  = load_model(model_choice) # Possibly pass in max_token_length as parameter into config

            checkpoint_path, patience_value = prepare_and_clean_data(dataset, learning_rate_choices, chosen_learning_rate, model_choice, 
                                                                    number_of_runs, validation_set_scoring, 
                                                                    label_column, validation_set, patience_value, 
                                                                    num_epochs, num_warmup_steps, gradient_accumulation_multiplier, 
                                                                    assigned_batch_size, tokenizer)

            synth_queries = analyze_and_report_data(dataset, label_column, tokenizer, max_token_length)

            train_df, test_set = transform_data(synth_queries, validation_set, label_column)

            train_set_text, train_set_label, dev_set_text, dev_set_label, test_set_text, text_set_label_, labels_list = split_dataset(train_df, dataset, test_set, label_column)

            training_dataset_pandas, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow, test_dataset_pandas = prepare_dataset(validation_set_scoring, train_set_label, train_set_text, dev_set_label, dev_set_text)

            tokenized_datasets = initalize_dataset_for_tokenization(tokenizer, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow)

            model, avg_train_losses, avg_valid_losses, eval_dataloader, inference_times = train_and_evaluate_model(number_of_runs, tokenized_datasets, assigned_batch_size, train_set_label, model_choice, chosen_learning_rate, 
                                    device, checkpoint_path, patience_value, num_epochs, num_warmup_steps, gradient_accumulation_multiplier)

            total_predictions, total_references, metric = evaluate_model(model, model_choice, checkpoint_path, device, eval_dataloader, inference_times)

            print_and_save_model(total_predictions, total_references, checkpoint_path, metric)





# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import load_model
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import prepare_and_clean_data
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import analyze_and_report_data
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import transform_data
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import split_dataset
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import prepare_dataset
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import initalize_dataset_for_tokenization
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import train_and_evaluate_model
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import evaluate_model 
# from ares.LLM_as_a_Judge_Adaptation.General_Binary_Classifier import print_and_save_model
# import pandas as pd
# import torch

# def binary_classifer_config(classification_datasets: list, validation_set: str, label_column: str, 
#                     num_epochs: int, patience_value: int, learning_rate: float, 
#                     model_choice: str = "microsoft/deberta-v3-large", validation_set_scoring: bool = True, 
#                     assigned_batch_size: int = 1, gradient_accumulation_multiplier: int = 32, 
#                     number_of_runs: int = 1, num_warmup_steps: int = 100):

#     device = "cuda:0"
#     device = torch.device(device)
#     learning_rate_choices = [learning_rate]
#     combined_dataset = None

#     if len(classification_datasets) == 1:
#         combined_dataset_path = pd.read_csv(classification_datasets[0], on_bad_lines='warn') 
#         print("Single file")
#     else: 
#         for dataset_path in classification_datasets:
#             dataset = pd.read_csv(dataset_path, on_bad_lines='warn') 
#             if combined_dataset is None:
#                 combined_dataset = dataset
#             else:
#                 combined_dataset = pd.concat([combined_dataset, dataset], ignore_index=True)

#         combined_dataset_path = "combined_classification_dataset.csv"
#         combined_dataset.to_csv(combined_dataset_path, index=False)

#     for chosen_learning_rate in learning_rate_choices:
#         tokenizer, max_token_length  = load_model(model_choice) # Possibly pass in max_token_length as parameter into config
        
#         checkpoint_path, patience_value = prepare_and_clean_data(combined_dataset_path, learning_rate_choices, chosen_learning_rate, model_choice, 
#                                                                 number_of_runs, validation_set_scoring, 
#                                                                 label_column, validation_set, patience_value, 
#                                                                 num_epochs, num_warmup_steps, gradient_accumulation_multiplier, 
#                                                                 assigned_batch_size, tokenizer)
        
#         synth_queries = analyze_and_report_data(combined_dataset_path, label_column, tokenizer, max_token_length)
        
#         train_df, test_set = transform_data(synth_queries, validation_set, label_column)
        
#         train_set_text, train_set_label, dev_set_text, dev_set_label, test_set_text, text_set_label_, labels_list = split_dataset(train_df, combined_dataset_path, test_set, label_column)
        
#         training_dataset_pandas, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow, test_dataset_pandas = prepare_dataset(validation_set_scoring, train_set_label, train_set_text, dev_set_label, dev_set_text)
        
#         tokenized_datasets = initalize_dataset_for_tokenization(tokenizer, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow)
        
#         model, avg_train_losses, avg_valid_losses, eval_dataloader, inference_times = train_and_evaluate_model(number_of_runs, tokenized_datasets, assigned_batch_size, train_set_label, model_choice, chosen_learning_rate, 
#                                 device, checkpoint_path, patience_value, num_epochs, num_warmup_steps, gradient_accumulation_multiplier)

#         total_predictions, total_references, metric = evaluate_model(model, model_choice, checkpoint_path, device, eval_dataloader, inference_times)

#         print_and_save_model(total_predictions, total_references, checkpoint_path, metric)






